#! /usr/local/bin/gawk -f
### ====================================================================
###  @Awk-file{
###     author          = "Nelson H. F. Beebe",
###     version         = "1.00",
###     date            = "02 July 1996",
###     time            = "07:56:44 MDT",
###     filename        = "epsutil.awk",
###     address         = "Center for Scientific Computing
###                        Department of Mathematics
###                        University of Utah
###                        Salt Lake City, UT 84112
###                        USA",
###     telephone       = "+1 801 581 5254",
###     FAX             = "+1 801 581 4148",
###     URL             = "http://www.math.utah.edu/~beebe",
###     checksum        = "57907 1524 6268 52617",
###     email           = "beebe@math.utah.edu (Internet)",
###     codesource      = "ftp://ftp.math.utah.edu/pub/misc/index.html#epsutil-1.00",
###     codetable       = "ISO/ASCII",
###     keywords        = "background; clipping; color; EPS;
###                        Encapsulated PostScript; framing; landscape;
###                        linewidth; portrait; PostScript utility;
###                        rotation; scaling; translation",
###     supported       = "yes",
###     abstract        = "This program inserts a wrapper around an
###                        Encapsulated PostScript file to accomplish a
###                        rotation down to the right by -90 degrees,
###                        turning landscape orientation into portrait
###                        orientation.  Command-line options are
###                        available for control of background color,
###                        clipping, linewidth, scaling, and
###                        translation.",
###     docstring       = "Filter an Encapsulated PostScript file, or
###                        files, named on the command line (or stdin if
###                        no filenames are given), to stdout, inserting
###                        wrapper PostScript to modify the image.
###
###                        Command-line options provide for background
###                        color, BoundingBox adjustment and
###                        measurement, checking for non-EPS operators,
###                        clipping, framing, linewidth scaling,
###                        PostScript alteration, rotation, scaling, and
###                        translation.
###
###                        Usage:
###                            nawk -f epsutil.awk \
###                                 [ BG=color-name ] \
###                                 [ BOUNDINGBOX=llx,lly,urx,ury ] \
###                                 [ CHECK=check ] \
###                                 [ CLIP=clip ] \
###                                 [ CMYK=cyan,magenta,yellow,black ] \
###                                 [ FRAME=frame-thickness ] \
###                                 [ GRAY=gray ] \
###                                 [ GRID=grid ] \
###                                 [ HEIGHT=height ] \
###                                 [ HSB=hue,saturation,brightness ] \
###                                 [ ID=idx,idy[,fontsize[,fontname]] ] \
###                                 [ LINEWIDTHSCALE=line-width-scale ] \
###                                 [ MAXLINEWIDTH=maximum-line-width ] \
###                                 [ MINLINEWIDTH=minimum-line-width ] \
###                                 [ OXY=origin ] \
###                                 [ OX=llx-origin ] \
###                                 [ OY=lly-origin ] \
###                                 [ PSCODE='PostScript code'] \
###                                 [ PSFILE=psfile,psfile,... ] \
###                                 [ RGB=red,green,blue ] \
###                                 [ RGBFILE=rgbfile ] \
###                                 [ ROTATE=degrees ] \
###                                 [ SXY=scale-factor ] \
###                                 [ SX=x-scale-factor ] \
###                                 [ SY=y-scale-factor ] \
###                                 [ WIDTH=width ] \
###                                 [ <] old-PS-file >new-PS-file
###
###                        The output file fully conforms to the
###                        PostScript Document Structuring Conventions
###                        defined in Appendix G of the PostScript
###                        Language Reference Manual, Second Edition,
###                        Addison-Wesley (1985), ISBN 0-201-10174-2
###                        (the Adobe red book).
###
###                        The output of this program is also suitable
###                        input to it; after four filtering steps with
###                        90-degree rotation selected, the original
###                        figure will be recovered, although the
###                        PostScript file will be about 4300 bytes
###                        larger for each 90-degree rotation because of
###                        the wrapper overhead.
###
###                        The original command line that was used to
###                        create the output file is preserved in a
###                        %%Title comment in the second line of the
###                        output file, as a record of the requested
###                        transformations.
###
###                        A verbatim copy of the original file, minus
###                        any %%EOF comments, is embedded in the output
###                        between the %%BeginDocument and %%EndDocument
###                        lines, where it can be recovered by simply
###                        stripping the wrapping PostScript.
###
###                        In the variables described below, if no
###                        dimension unit is specified, physical
###                        dimensions are measured in standard
###                        PostScript units of big points:
###
###                                72bp == 1in == 25.4mm
###
###                        Otherwise, any of the TeX dimension units of
###                        bp (big point), in (inch), cm (centimeter),
###                        pt (point), cc (cicero), dd (didot point), pc
###                        (pica), or sp (scaled point), may suffix a
###                        number, and the value will be converted to
###                        big points internally.
###
###
###                        ================
###                        Background color
###                        ================
###
###                        A background color can be specified as a
###                        named color, or as a triple (or, for CYMK,
###                        quadruple) of numbers, each in the range
###                        0..1.  The numbers must separated by any
###                        character other than a digit or period.
###                        Comma or slash is a readable choice for a
###                        separator character, and avoids the need to
###                        quote the values to protect them from
###                        possible shell interpretation.
###
###                        Five color models are available.  The models
###                        are tried in alphabetical order: the first
###                        one specified is used, and any settings for
###                        the others are silently ignored.
###
###                        For BG, the value is the name of a color
###                        found in either the file specified by the
###                        RGBFILE=rgbfile option, or else in the X
###                        Window System rgb.txt color mapping
###                        file. These files associate red, green, and
###                        blue intensities with named colors; their
###                        lines contain three numbers in the range
###                        0..255, followed by a color name.  The
###                        rgb.txt file is found by searching a built-in
###                        directory path that covers all major UNIX
###                        systems; that path can be overridden by
###                        defining a value for the environment variable
###                        RGBPATH.  For example, on an IBM RS/6000 AIX
###                        system, you could set that variable to
###                        /usr/lpp/X11/lib/X11/It_IT to get Italian
###                        color names.
###
###                        In CMYK, values for cyan, magenta, and yellow
###                        are provided (0 means no color, 1 means full
###                        color), and then an amount of black (0 means
###                        none, 1 means completely black) is added to
###                        darken the color.
###
###                        In GRAY, 0 is black and 1 is white. GRAY=g is
###                        equivalent to RGB=g,g,g.
###
###                        In HSB, hue runs through the rainbow
###                        (approximately, red=0, orange=0.1,
###                        limegreen=0.2, green=0.3, cyan=0.5, blue=0.6,
###                        magenta=0.8, deeppink=0.9, red=1). Saturation
###                        goes from 0 (no color == black) to 1 (full
###                        color).  Brightness goes from 0 (dark) to 1
###                        (bright).
###
###                        In RGB, major colors are:
###
###                              Black   0,0,0
###                              Red     1,0,0
###                              Green   0,1,0
###                              Yellow  1,1,0
###                              Blue    0,0,1
###                              Magenta 1,0,1
###                              Cyan    0,1,1
###                              White   1,1,1
###
###
###                        ===========
###                        BoundingBox
###                        ===========
###
###                        Encapsulated PostScript files are required to
###                        contain a %%BoundingBox comment that defines
###                        the lower-left and upper-right corner
###                        coordinates in big points.  Sadly, most
###                        PostScript-producing programs produce
###                        (sometimes wildly) incorrect coordinate
###                        values in this comment, so manual adjustment
###                        is frequently necessary.  Also, sometimes
###                        only a portion of the original figure is
###                        required, necessitating definition of a new
###                        BoundingBox.
###
###                        To avoid the need for modifying the original
###                        PostScript file, the two corner coordinate
###                        pairs can be specified as four
###                        optionally-dimensioned numbers for the value
###                        of the BOUNDINGBOX option.  Dimensionless
###                        numbers are assumed to be measured in big
###                        points.  The numbers can be separated by any
###                        characters other than plus, minus, digits,
###                        period, or lowercase letters; comma or slash
###                        are readable choices.  Values specified this
###                        way override those in the %%BoundingBox
###                        comment.
###
###                        When a BOUNDINGBOX value is given, especially
###                        if it covers only a portion of the original
###                        figure, clipping should be selected.
###
###                        Although fractional dimensions are accepted,
###                        after converting to big points, the values
###                        actually used will be rounded down to the
###                        nearest integer at the lower-left corner, and
###                        up to the nearest integer at the upper-right
###                        corner.  This is a requirement of the
###                        PostScript Document Structuring Conventions.
###
###
###                        =================
###                        BoundingBox frame
###                        =================
###
###                        A rectangular frame of a specified line
###                        thickness will be drawn around the outside of
###                        the original BoundingBox if the
###                        FRAME=frame-thickness option is given.
###
###                        If linewidth scaling is requested, the frame
###                        will also be affected by the scaling.
###
###                        The output BoundingBox will automatically be
###                        enlarged by the twice the frame thickness,
###                        and any MAXLINEWIDTH value will be adjusted
###                        upward if necessary to ensure that the frame
###                        can be drawn properly.
###
###
###                        ===============================
###                        Checking for non-EPSF operators
###                        ===============================
###
###                        If a non-zero value of the CHECK=check
###                        option is given, check for the presence
###                        of PostScript operators that are not
###                        permitted in EPS files.
###
###                        Because this option requires additional
###                        pattern matching on every input line, and
###                        is only occasionally needed, the default is
###                        CHECK=0 to suppress the check.
###
###
###                        ========
###                        Clipping
###                        ========
###
###                        If a zero value of the CLIP=clip option is
###                        given, suppress clipping the figure to the
###                        BoundingBox.
###
###                        Although EPS-including programs are supposed
###                        to provide this clipping automatically, not
###                        all do so.  Thus, the default in this program
###                        is to turn on clipping.
###
###                        The clipping path is set after geometric
###                        transformations have been completed, but
###                        before any drawing operations are performed.
###                        Thus, the BoundingBox frame, and the
###                        background color, are subject to clipping.
###
###
###                        ====
###                        Grid
###                        ====
###
###                        A non-zero value for the GRID=grid option
###                        requests that a grid be drawn over the
###                        picture to allow convenient determination of
###                        BoundingBox coordinates for a portion of the
###                        view.
###
###                        Clipping is automatically suppressed when a
###                        grid is selected, in order for the complete
###                        grid, with its edge labels, to be visible.
###
###
###                        =================
###                        Identifying label
###                        =================
###
###                        When working with large numbers of figures,
###                        or experimenting with various options, it is
###                        sometimes difficult to keep track of them.
###                        The ID option can be specified to provide a
###                        page location where an identifying label
###                        recording the command line will be printed.
###                        The label location is measured on the output
###                        page, relative to the lower-left corner,
###                        independent of geometric transformations.  It
###                        is drawn before any background fill color is
###                        supplied, and will be overwritten by that
###                        fill if it lies within the BoundingBox of the
###                        output figure.
###
###                        An optional third value defines the font size
###                        (default: 10bp), and an optional fourth value
###                        defines the font name (default: Helvetica).
###
###                        Values are separated by characters other than
###                        plus, minus, digits, period, or letters.
###
###
###                        =================
###                        Linewidth control
###                        =================
###
###                        Sometimes, linewidths need adjustment, usually
###                        to thicken them for better reproduction.  The
###                        LINEWIDTHSCALE value causes all linewidths to
###                        be multiplied by that value.  This scaling
###                        also affects any framing rectangle that is
###                        requested.
###
###                        PostScript allows zero linewidths, which
###                        means the thinnest line that the output
###                        device is capable of producing.  Such
###                        linewidths would not be affected by scaling.
###                        Thus, a value for MINLINEWIDTH can be given
###                        to force all lines to be at least that thick,
###                        and therefore, subject to scaling.
###
###                        To prevent scaling from producing
###                        ridiculously thick lines, a MAXLINEWIDTH
###                        value can be set to force an upper limit to
###                        linewidths.  If not specified, it defaults to
###                        24.
###
###
###                        =====================
###                        PostScript insertions
###                        =====================
###
###                        Two options, PSCODE and PSFILE, provide for
###                        insertion of additional PostScript code into
###                        the end of the %%BeginProlog ...  %%EndProlog
###                        section where PostScript commands are
###                        defined.
###
###                        The value of PSCODE is a string containing
###                        arbitrary literal PostScript code.
###
###                        The value of PSFILE is the name of one or
###                        more comma- or space-separated files whose
###                        contents are to be inserted.
###
###                        When both options are specified, PSFILE is
###                        handled before PSCODE, so that any PostScript
###                        read from files can be overridden by a
###                        literal PostScript from the command line.
###
###
###                        ========
###                        Rotation
###                        ========
###
###                        The ROTATE option value is a rotation angle
###                        in degrees, positive counterclockwise, and
###                        negative clockwise.  Only multiples of +/-90
###                        degrees are supported.
###
###                        For example, to turn landscape orientation
###                        with the right side at the top of the page
###                        into portrait orientation, specify
###                        ROTATE=-90.  It will usually be necessary to
###                        supply additional scaling (HEIGHT or WIDTH,
###                        or SX, SY, SXY options) to make the picture
###                        fit on the output page.
###
###                        Unless otherwise specified by the translation
###                        variables (see below), the position of the
###                        lower-left corner of the BoundingBox on the
###                        output display device is kept invariant, so
###                        that after a -90-degree rotation, the
###                        lower-right corner of the original
###                        BoundingBox will be positioned where the
###                        lower-left corner was.
###
###
###                        ===========
###                        Translation
###                        ===========
###
###                        The origin variables move the lower-left
###                        corner of the bounding box to (OX,OY), or if
###                        that is not specified, then to (OXY,OXY).
###                        Otherwise, the lower-left corner of the new
###                        bounding box is identical to the original.
###
###
###                        ==================
###                        Scaling (Absolute)
###                        ==================
###
###                        The HEIGHT=height and WIDTH=width options
###                        provide a way to request absolute scaling
###                        of the BoundingBox to a specified height or
###                        width.  If only one of them is given, then
###                        the other is computed to preserve the
###                        aspect ratio for distortion-free scaling.
###                        If both are given, then the horizontal and
###                        vertical scale factors may be different; the
###                        aspect ratio will then not be preserved.
###
###                        The two sizes refer to the final figure,
###                        after rotation.
###
###                        If both absolute (HEIGHT and WIDTH) and
###                        relative (SX, SY, and SXY) scaling are
###                        specified, then absolute scaling is done,
###                        and the relative scale factors are ignored.
###
###
###                        ==================
###                        Scaling (Relative)
###                        ==================
###
###                        The SX, SY, and SXY options provide for
###                        relative scaling.  Any scale factors
###                        provided are applied AFTER the rotation, so
###                        that they apply to the final figure. SX and
###                        SY are used if provided; otherwise, SXY is
###                        used for both x and y scaling if provided;
###                        otherwise, no scaling is done.
###
###
###                        =================
###                        Hand Modification
###                        =================
###
###                        The output file can be easily modified by
###                        hand later to adjust any of the options that
###                        can be set on the command line.  Look for the
###                        lines following the first %%BeginProlog
###                        comment that begin /NAME ... def, and then
###                        change the numeric values as needed.  Comment
###                        lines precede each set of related variable
###                        definitions to provide suitable
###                        documentation.
###
###
###                        =============
###                        Memory limits
###                        =============
###
###                        Until a valid %%BoundingBox comment has been
###                        located, or a BOUNDINGBOX command-line option
###                        has been supplied, the input PostScript is
###                        buffered in memory.  On small machines, or
###                        with very large PostScript files that contain
###                        the %%BoundingBox comment near the end
###                        instead of near the beginning, it is possible
###                        that your computer memory (and swap space, if
###                        any), might be exceeded.  The easiest
###                        solution for such a problem is to supply a
###                        command-line BOUNDINGBOX value.  This problem
###                        is expected to be exceedingly rare.
###
###
###                        =====
###                        TO DO
###                        =====
###
###                        Maybe add color remapping.
###
###
###                        ========
###                        Checksum
###                        ========
###
###                        The checksum field above contains a CRC-16
###                        checksum as the first value, followed by the
###                        equivalent of the standard UNIX wc (word
###                        count) utility output of lines, words, and
###                        characters.  This is produced by Robert
###                        Solovay's checksum utility.",
###  }
### ====================================================================

BEGIN           { pre_initialize() }

/^%!/		{ if (pre_scan) initialize() }

/^%%EOF/        { next }        # discard EOF comments

/^%%BoundingBox: *[0-9.]+[ \t]+[0-9.]+[ \t]+[0-9.]+[ \t]+[0-9.]+/ \
		{ if (pre_scan) BoundingBox($2,$3,$4,$5) }

		{ output($0) }  # copy everything verbatim

END             { end_wrapper() }

#=======================================================================

function BoundingBox(llx,lly,urx,ury, k)
{
    LLX = floor(llx - frame)
    LLY = floor(lly - frame)
    URX = ceil(urx + frame)
    URY = ceil(ury + frame)
    start_wrapper()
}


function bigpoint(dimension)
{
    return (72.0 * inch(dimension))
}


function ceil(x)
{
    return ((int(x) < x) ? (int(x) + 1) : int(x))
}


function check_EPS_operators(s, k)
{
    k = index(s,"%")
    if (k == 0)
	k = length(s)
    if ((k > 1) && match(substr(s,1,k), \
	"banddevice|grestoreall|nulldevice|setpageparams|copypage|" \
	"initclip|quit|setsccbatch|erasepage|initgraphics|renderbands|" \
	"setscreen|exitserver|initmatrix|setdevice|settransfer|framedevice|" \
	"note|setmatrix"))
	warning("operator [" substr(s,RSTART,RLENGTH) "] is illegal in EPS file")
}


function do_header()
{
    do_comment_header()
    do_prolog()
    do_setup()
    do_page_setup()
}


function define_clipping_path()
{
    print "% Step 5: Set clipping path to ORIGINAL BoundingBox"
    print "newpath"
    print "    LLX LLY moveto"
    print "    LLX URY lineto"
    print "    URX URY lineto"
    print "    URX LLY lineto"
    print "closepath"
    print "CLIP { clip } if"
    print ""
}


function define_EPSF_wrappers()
{
    # These definitions are from p. 726 of the Adobe
    # PostScript Language Reference Manual
    print "% Include file wrapper macros from p. 726 of Adobe red book"
    print "/BeginEPSF"
    print "{"
    print "    /b4_Inc_state save def"
    print "    /dict_count countdictstack def"
    print "    /op_count count 1 sub def"
    print "    userdict begin"
    print "    /showpage { } def"
    print "    0 setgray"
    print "    0 setlinecap"
    print "    1 setlinewidth"
    print "    0 setlinejoin"
    print "    10 setmiterlimit"
    print "    [ ] 0 setdash"
    print "    newpath"
    print "    /languagelevel where"
    print "    {"
    print "        pop languagelevel"
    print "        1 ne "
    print "        {"
    print "            false setstrokeadjust"
    print "            false setoverprint"
    print "        } if"
    print "    } if"
    print "} bind def"
    print ""
    print "/EndEPSF"
    print "{"
    print "    count op_count sub {pop} repeat"
    print "    countdictstack dict_count sub {end} repeat"
    print "    b4_Inc_state restore"
    print "} bind def"
    print ""
}


function define_grid()
{
    print "%%BeginResource: procset gridmacros 1.01"
    print "/griddict 60 dict def"
    print "griddict begin"
    print "/TickSize 5 def"
    print "/TickStep 5 def"
    print "/TickFontSize 9 def"
    print ""
    print "/Frame                                  % x y width height Frame"
    print "{"
    print "        /height exch def"
    print "        /width exch def"
    print "        /y exch def"
    print "        /x exch def"
    print "        newpath"
    print "        x y moveto"
    print "        width 0 rlineto"
    print "        0 height rlineto"
    print "        width neg 0 rlineto"
    print "        closepath"
    print "        stroke"
    print ""
    print "        x y width TickStep TickSize TickX"
    print "        x y width TickStep TickSize NumberX"
    print "        x y height add width TickStep TickSize neg TickX"
    print "        x y height add width TickStep TickSize neg NumberX"
    print "        x y height TickStep TickSize TickY"
    print "        x y height TickStep TickSize NumberY"
    print "        x width add y height TickStep TickSize neg TickY"
    print "        x width add y height TickStep TickSize neg NumberY"
    print "        x y 25 25 width height Grid"
    print "} def"
    print ""
    print "/Grid                                   % x y dx dy width height Grid"
    print "{"
    print "        /Gheight exch def"
    print "        /Gwidth exch def"
    print "        /Gdy exch def"
    print "        /Gdx exch def"
    print "        /Gy exch def"
    print "        /Gx exch def"
    print ""
    print "        Gx"
    print "        Gdx"
    print "        Gx Gwidth add"
    print "        {"
    print "                newpath"
    print "                Gy moveto"
    print "                0 Gheight rlineto"
    print "                closepath"
    print "                stroke"
    print "        }"
    print "        for"
    print ""
    print "        Gy"
    print "        Gdy"
    print "        Gy Gheight add"
    print "        {"
    print "                newpath"
    print "                Gx exch moveto"
    print "                Gwidth 0 rlineto"
    print "                closepath"
    print "                stroke"
    print "        }"
    print "        for"
    print "} def"
    print ""
    print ""
    print "/NumberX                                % x y length dx tick NumberX"
    print "{"
    print "        /Helvetica findfont TickFontSize scalefont setfont"
    print "        /Ttick exch def"
    print "        /Tdx exch def"
    print "        /Tlength exch def"
    print "        /Ty exch def"
    print "        /Tx exch def"
    print ""
    print "        /temp 12 string def"
    print "        Tx"
    print "        Tdx 5 mul"
    print "        Tx Tlength add"
    print "        {"
    print "                dup"
    print "                Ty moveto               % Tx already on stack"
    print "                Ttick 0 lt"
    print "                { 0 TickFontSize 0.5 mul rmoveto }"
    print "                { 0 TickFontSize -1.5 mul rmoveto }"
    print "                ifelse"
    print "                temp cvs"
    print "                dup stringwidth pop -2 div 0 rmoveto"
    print "                show"
    print "        }"
    print "        for"
    print "} def"
    print ""
    print "/NumberY                                % x y length dy tick NumberY"
    print "{"
    print "        /Helvetica findfont TickFontSize scalefont setfont"
    print "        /Ttick exch def"
    print "        /Tdy exch def"
    print "        /Tlength exch def"
    print "        /Ty exch def"
    print "        /Tx exch def"
    print ""
    print "        /temp 12 string def"
    print "        Ty"
    print "        Tdy 5 mul"
    print "        Ty Tlength add"
    print "        {"
    print "                dup"
    print "                Tx exch moveto          % Ty already on stack"
    print "                Ttick neg TickFontSize -2 div rmoveto"
    print "                temp cvs"
    print "                Ttick 0 gt"
    print "                { dup stringwidth pop neg 0 rmoveto }"
    print "                if"
    print "                show"
    print "        }"
    print "        for"
    print "} def"
    print ""
    print "/TickX                                  % x y length dx tick TickX"
    print "{"
    print "        /Ttick exch def"
    print "        /Tdx exch def"
    print "        /Tlength exch def"
    print "        /Ty exch def"
    print "        /Tx exch def"
    print ""
    print "        Tx"
    print "        Tdx"
    print "        Tx Tlength add"
    print "        {"
    print "                dup"
    print "                newpath"
    print "                Ty moveto               % Tx already on stack"
    print "                cvi Tdx 5 mul cvi mod 0 eq% draw tick, double length every 5th"
    print "                { Ttick 2 mul }"
    print "                { Ttick }"
    print "                ifelse"
    print "                0"
    print "                exch"
    print "                rlineto"
    print "                closepath"
    print "                stroke"
    print "        }"
    print "        for"
    print "} def"
    print ""
    print "/TickY                                  % x y length dy tick TickY"
    print "{"
    print "        /Ttick exch def"
    print "        /Tdy exch def"
    print "        /Tlength exch def"
    print "        /Ty exch def"
    print "        /Tx exch def"
    print ""
    print "        Ty"
    print "        Tdy"
    print "        Ty Tlength add"
    print "        {"
    print "                dup"
    print "                newpath"
    print "                Tx exch moveto          % Ty already on stack"
    print "                cvi Tdy 5 mul cvi mod 0 eq% draw tick, double length every 5th"
    print "                { Ttick 2 mul }"
    print "                { Ttick }"
    print "                ifelse"
    print "                0"
    print "                rlineto"
    print "                closepath"
    print "                stroke"
    print "        }"
    print "        for"
    print "} def"
    print ""
    print "/inch"
    print "{"
    print "        72 mul"
    print "} def"
    print ""
    print "/mm"
    print "{"
    print "        72 mul 25.4 div"
    print "} def"
    print ""
    print "% (string) CenterCenterShow -- show centered at currentpoint"
    print "/CenterCenterShow"
    print "{"
    print "        dup stringwidth pop -2 div TickFontSize -2 div rmoveto show"
    print "} def"
    print ""
    print "% (string) RightCenterShow -- show right-centered at currentpoint"
    print "/RightCenterShow"
    print "{"
    print "        dup stringwidth pop neg TickFontSize -2 div rmoveto show"
    print "} def"
    print ""
    print "% (string) LeftCenterShow -- show right-centered at currentpoint"
    print "/LeftCenterShow"
    print "{"
    print "        0 TickFontSize -2 div rmoveto show"
    print "} def"
    print ""
    print "% Select ONE of these pairs of page dimensions"
    print ""
    print "% ISO A4-format paper"
    print "% /PageWidth   210 mm def"
    print "% /PageHeight  297 mm def"
    print ""
    print "% US A-format paper"
    print "/PageWidth   8.5 inch def"
    print "/PageHeight 11.0 inch def"
    print "end % griddict"
    print ""
    print "%%EndProcset"
    print ""
}

function define_id( n,words)
{
    n = split(ID,words,"[^0-9.A-Za-z+\-]")

    if (n < 1)
        words[1] = "30"
    if (n < 2)
        words[2] = "740"
    if (n < 3)
        words[3] = "10"
    if (n < 4)
        words[4] = "Helvetica"

    print "/" words[4], "findfont", bigpoint(words[3]),"scalefont setfont"
    print "0 setgray"
    print bigpoint(words[1]),bigpoint(words[2]),"moveto"
    print "(" command_line ")"
    print "ID { show } { pop } ifelse"
}


function define_one_psfile(filename, line)
{
    print "%%BeginResource: file",filename
    while ((getline line < filename) > 0)
	print line
    print "%%EndResource"
    close(filename)
}


function define_pscode()
{
    if (PSCODE != "")
    {
	print "% User-defined PSCODE"
        print PSCODE
    }
}


function define_psfile( k,n,filename)
{
    if (PSFILE != "")
    {
	n = split(PSFILE,filename,"[, \t]")
	for (k = 1; k <= n; ++k)
	{
	    sub(/^[ \t]*/,"",filename[k]) # trim leading space
	    sub(/[ \t]*$/,"",filename[k]) # trim trailing space
	    define_one_psfile(filename[k])
	}
    }
}


function define_rectfill()
{
    define_rectxxx("fill")
}


function define_rectstroke()
{
    define_rectxxx("stroke")
}


function define_rectxxx(operator)
{
    print "% Define a Level 2 PostScript operator in case we are in Level 1"
    print "/rect" operator
    print "{"
    print "    /H exch def"
    print "    /W exch def"
    print "    /Y exch def"
    print "    /X exch def"
    print "    gsave"
    print "        newpath"
    print "            X Y moveto"
    print "            W 0 rlineto"
    print "            0 H rlineto"
    print "            W neg 0 rlineto"
    print "            closepath"
    print "       ", operator
    print "    grestore"
    print "} def"
    print ""
}


function define_setlinewidth()
{
    print "% Linewidth control"
    print "/MINLINEWIDTH", minlinewidth, "def"
    print "/MAXLINEWIDTH", maxlinewidth, "def"
    print "/LINEWIDTHSCALE", linewidthscale, "def"
    print "/old_setlinewidth systemdict begin /setlinewidth load end def"
    print "/setlinewidth"
    print "{"
    print "    LINEWIDTHSCALE mul"
    print "    dup MINLINEWIDTH lt { pop MINLINEWIDTH } if"
    print "    dup MAXLINEWIDTH gt { pop MAXLINEWIDTH } if"
    print "    old_setlinewidth"
    print "} bind def"
    print ""
}


function define_transformation( degrees)
{
    print "% Step 1: Translate the ORIGINAL figure to place its lower-left corner"
    print "%         at the new page position (OX, OY)"
    print "LLX OX sub neg LLY OY sub neg translate"
    print ""

    degrees = (ROTATE == "") ? 0 : ((360 + (ROTATE % 360)) % 360)

    if (degrees == 0)
	define_rotate_000()
    else if (degrees == 90)
	define_rotate_090()
    else if (degrees == 180)
	define_rotate_180()
    else if (degrees == 270)
	define_rotate_270()
    else
	warning("rotation angle must be a multiple of +/-90 degrees: saw [" ROTATE "]")
}

function define_rotate_000()
{
    print "% Step 2: No rotation needed"
    print ""

    print "% Step 3: No translation needed"
    print ""

    print "% Step 4: Scale the drawing about the ORIGINAL lower-left corner"
    print "LLX LLY translate"
    print "SX SY scale"
    print "LLX neg LLY neg translate"
    print ""
}


function define_rotate_090()
{
    print "% Step 2: Rotate 90 degrees up about the lower-left BoundingBox corner"
    print "LLX LLY translate"
    print "90 rotate"
    print "LLX neg LLY neg translate"
    print ""

    print "% Step 3: Translate the ORIGINAL upper-left corner to"
    print "%         the ORIGINAL lower-left corner"
    print "0 URY LLY sub neg translate"
    print ""

    print "% Step 4: Scale the drawing about the ORIGINAL upper-left corner"
    print "LLX URY translate"
    print "SX SY scale"
    print "LLX neg URY neg translate"
    print ""
}


function define_rotate_180()
{
    print "% Step 2: Rotate 180 degrees up about the lower-left BoundingBox corner"
    print "LLX LLY translate"
    print "180 rotate"
    print "LLX neg LLY neg translate"
    print ""

    print "% Step 3: Translate the ORIGINAL upper-right corner to"
    print "%         the ORIGINAL lower-left corner"
    print "URX LLX sub neg URY LLY sub neg translate"
    print ""

    print "% Step 4: Scale the drawing about the ORIGINAL upper-right corner"
    print "URX URY translate"
    print "SX SY scale"
    print "URX neg URY neg translate"
    print ""
}


function define_rotate_270()
{
    print "% Step 2: Rotate 90 degrees down about the lower-left BoundingBox corner"
    print "LLX LLY translate"
    print "-90 rotate"
    print "LLX neg LLY neg translate"
    print ""

    print "% Step 3: Translate the ORIGINAL lower-right corner to"
    print "%         the ORIGINAL lower-left corner"
    print "URX LLX sub neg 0 translate"
    print ""

    print "% Step 4: Scale the drawing about the ORIGINAL lower-right corner"
    print "URX LLY translate"
    print "SX SY scale"
    print "URX neg LLY neg translate"
    print ""
}


function define_user_variables()
{
    print "% Define the clipping option"
    print "/CLIP",(CLIP ? "true" : "false"),"def"
    print ""

    print "% Define BoundingBox frame linewidth (negative value suppresses)"
    print "/FRAME",((FRAME == "") ? "-1" : bigpoint(FRAME)),"def"
    print ""

    print "% Define the identifying label option"
    print "/ID",((ID == "") ? "false" : "true"),"def"
    print ""

    print "% Define the ORIGINAL BoundingBox coordinates"
    print "% Units: PostScript big points: 72bp == 1in == 25.4mm"
    print "/LLX", LLX, "def"
    print "/LLY", LLY, "def"
    print "/URX", URX, "def"
    print "/URY", URY, "def"
    print ""

    print "% Define the new position of the lower-left BoundingBox corner"
    print "% Units: PostScript big points: 72bp == 1in == 25.4mm"
    print "/OX", OX, "def"
    print "/OY", OY, "def"
    print ""

    print "% Define the x and y scale factors"
    print "/SX", SX, "def"
    print "/SY", SY, "def"
    print ""
}


function do_comment_header( degrees)
{
    print "%!PS-Adobe-3.0 EPSF-3.0"

    command_line = "/usr/local/bin/gawk -f epsutil.awk"
    for (k = 1; k < ARGC; ++k)
    {
	if (ARGV[k] ~ /[\[\]\{\} *?"]/)
	    command_line = command_line " '" ARGV[k] "'"
	else
	    command_line = command_line " " ARGV[k]
    }
    print "%%Title:", command_line

    print "%%CreationDate: Sat Jun 29 06:19:29 1996"
    print "%%Creator: epsutil.awk", VERSION
    print "%%For: " personal_name " <" username "@" hostname ">"

    degrees = (ROTATE == "") ? 0 : ((360 + (ROTATE % 360)) % 360)
    if ((degrees == 90) || (degrees == 270))
        print "%%BoundingBox:", floor(OX), floor(OY), \
		ceil(OX + SY*(URY - LLY)), ceil(OY + SX*(URX - LLX))
    else # (degrees == 0) || (degrees == 180)
        print "%%BoundingBox:", floor(OX), floor(OY), \
		ceil(OX + SX*(URX - LLX)), ceil(OY + SY*(URY - LLY))

    print "%%DocumentMedia: A 612 792 () () ()"
    print "%%DocumentData: Clean7Bit"
    print "%%DocumentFonts:"
    print "%%DocumentNeededResources:"
    print "%%DocumentSuppliedResources:"
    print "%%Orientation: Portrait"
    print "%%PageOrder: Ascend"
    print "%%Pages: 1"
    print "%%EndComments"
    print "%======================================================================="
}


function do_grid()
{
    print "%%BeginResource: procset gridshow 1.01"
    print "griddict begin"
    print ""
    print "% Make grid lines half the normal thickness"
    print "currentlinewidth 0.5 mul setlinewidth"
    print ""
    print "% You can change this line to make a different grid.  The values"
    print "% should all be multiples of 25."
    print "% The values 50 50 525 700 are appropriate for an Apple LaserWriter II."
    print "% x  y width height Frame"
    print "50 50 525 700 Frame"
    print ""
    print "% label bottom horizontal axis at each end"
    print "50 30 moveto"
    print "(bp) CenterCenterShow"
    print "575 30 moveto"
    print "(bp) CenterCenterShow"
    print ""
    print "% label top horizontal axis at each end"
    print "50 770 moveto"
    print "(bp) CenterCenterShow"
    print "575 770 moveto"
    print "(bp) CenterCenterShow"
    print ""
    print "% label left vertical axis at each end"
    print "25 50 moveto"
    print "(bp) RightCenterShow"
    print "25 750 moveto"
    print "(bp) RightCenterShow"
    print ""
    print "% label right vertical axis at each end"
    print "600 50 moveto"
    print "(bp) LeftCenterShow"
    print "600 750 moveto"
    print "(bp) LeftCenterShow"
    print "end % griddict"
    print "%%EndResource"
}


function do_page_setup()
{
    print "%%Page: 1 1"
    print "%%BeginPageSetup"
    print "rotatedict begin"

    define_id()

    define_transformation()

    define_clipping_path()

    if (BG != "")
	set_named_color()
    else if (CMYK != "")
	set_cmyk()
    else if (HSB != "")
	set_hsb()
    else if (GRAY != "")
	set_gray()
    else if (RGB != "")
	set_rgb()
    else
	set_color("RGB","1,1,1",3,"setrgbcolor")

    print "%%EndPageSetup"
}


function do_prolog()
{
    print "%%BeginProlog"
    print "/rotatedict 1000 dict def" # NB: large: private code may be supplied
    print "rotatedict begin"
    print ""

    define_user_variables()
    define_EPSF_wrappers()
    define_setlinewidth()
    define_rectfill()
    define_rectstroke()
    if (GRID)
	define_grid()
    define_psfile()		# NB: handle files before command-line code,
    define_pscode()		# so the latter can override the former

    print "/NormalState save def % remember normal state"
    print ""
    print "end % rotatedict"
    print "%%EndProlog"
    print "%======================================================================="
}


function do_setup()
{
    print "%%BeginSetup"
    print "/#copies 1 def"
    print "%%EndSetup"
    print "%======================================================================="
}


function do_trailer()
{
    print "end % rotatedict"
    print "showpage"
    print "%%PageTrailer"
    print "%======================================================================="
    print "%%Trailer"
    print "%%EOF"
}


function end_wrapper()
{
    if (pre_scan)               # no valid BoundingBox found, so provide
	BoundingBox(0,0,612,792) # a default one for US A-size paper

    print "%%EndDocument"
    print "EndEPSF"

    set_frame()			# must draw frame AFTER picture to ensure
				# overwriting of anything outside the
				# original BoundingBox

    print "% Return to normal graphics state"
    print "NormalState restore"
    print ""

    if (GRID)
	do_grid()

    do_trailer()
}


function file_exists(filename)
{
    # Unfortunately, getline in both nawk and gawk will hang if its
    # file does not exist, so we need to test for file existence by
    # invoking a shell command, sigh...

    return (system("test -f " filename) == 0)
}


function find_file(path,filename, fullname,k)
{
    for (k = 1; k in path; ++k)
    {
	fullname = path[k] "/" filename
	if (file_exists(fullname))
	    return (fullname)
    }
    return ("")
}


function floor(x)
{
     return ((int(x) > x) ? (int(x) - 1) : int(x))
}


function in_range(a,b,c)
{
    return ((a <= b) && (b <= c))
}


function inch(dimension, number,units)
{
    number = dimension + 0	# coerce to number, ignoring units

    if (match(dimension, \
	/^[ \t]*[-+]?[0-9]+[.]?[0-9]*(bp|in|cm|pt|cc|dd|mm|pc|sp)?[ \t]*$/))
    {
	units = substr(dimension,RSTART+RLENGTH-2,2)

	if (units == "bp")	# big points
	    number /= 72.0
	else if (units == "in")	# inches
	    ; # NO-OP
	else if (units == "cm")	# centimeters
	    number /= 2.54
	else if (units == "pt")	# points
	    number /= 72.27
	else if (units == "cc")	# cicero
	    number *= 12.0 * (1238.0 / 1157.0) / 72.27
	else if (units == "dd")	# didot points
	    number *= (1238.0 / 1157.0) / 72.27
	else if (units == "mm")	# millimeters
	    number /= 25.4
	else if (units == "pc")	# pica
	    number *= 12.0 / 72.27
	else if (units == "sp")	# scaled points
	    number /= (65536.0 * 72.27)
	else			# must be dimensionless number, assume bp
	    number /= 72.0
    }
    else
	warning("unrecognized dimension [" dimension "]: bp assumed")

    return (number)
}


function initialize( word)
{
    CHECK = (CHECK == "") ? 0 : (0 + CHECK)
    CLIP  = (CLIP == "")  ? 1 : (0 + CLIP)
    GRID  = (GRID == "")  ? 0 : (0 + GRID)

    if (GRID)			# must turn off clipping when grid is used
	CLIP = 0

    frame = (FRAME == "") ? 0 : bigpoint(FRAME)
    linewidthscale = (LINEWIDTHSCALE == "") ? 1 : (0 + LINEWIDTHSCALE)
    minlinewidth = bigpoint((MINLINEWIDTH == "") ? 0 : MINLINEWIDTH)
    maxlinewidth = bigpoint((MAXLINEWIDTH == "") ? 24 : MAXLINEWIDTH)
    if (maxlinewidth < frame)
	maxlinewidth = frame
    if (BOUNDINGBOX != "")
    {
	if (split(BOUNDINGBOX,word,"[^0-9.a-z+\-]") == 4)
	    BoundingBox(bigpoint(word[1]),bigpoint(word[2]), \
			bigpoint(word[3]),bigpoint(word[4]))
	else
	    warning("BOUNDINGBOX option needs 4 dimensions: saw [" BOUNDINGBOX "]")
    }
}


function output(s)
{
    if (CHECK)
	check_EPS_operators(s)
    if (pre_scan)
	lines[++nlines] = s
    else
	print s
}


function pre_initialize( word)
{
    # Change these two lines whenever the program is modified
    VERSION_NUMBER = "1.00"
    VERSION = "Version " VERSION_NUMBER " [02-Jul-1996]"

    "date" | getline current_date_and_time
    if ("USER" in ENVIRON)
	username = ENVIRON["USER"]
    else if ("LOGNAME" in ENVIRON)
	username = ENVIRON["LOGNAME"]
    else
	username = "unknown"
    "awk -F: '/^" username ":/ {print $5}' /etc/passwd" | \
	getline personal_name
    "hostname" | getline hostname

    nlines = 0
    pre_scan = 1
}


function set_cmyk()
{
    set_color("CMYK",CMYK,4,"setcmykcolor")
}


function set_color(name,value,count,operator, word)
{
    if ((split(value,word,"[^0-9.]+") == count) && \
	in_range(0,word[1],1) && \
	in_range(0,word[2],1) && \
	in_range(0,word[3],1) && \
	in_range(0,word[count],1))
    {
	print "% Fill background with solid color"
	print "gsave"
	if (count == 1)
	    print "    " word[1], operator
	else if (count == 3)
	    print "    " word[1], word[2], word[3], operator
	else if (count == 4)
	    print "    " word[1], word[2], word[3], word[4], operator
	else
	    warning("internal error: bad count [" count "] passed to set_color()")
	print "    LLX LLY URX LLX sub URY LLY sub rectfill"
	print "grestore"
	print ""
    }
    else
	warning(name " variable needs " count " number(s) in 0..1; saw [" value "]")
}


function set_frame()
{
    print "% Draw optional BoundingBox frame"
    print "FRAME 0 ge"
    print "{"
    print "    gsave"
    print "        0 setgray"
    print "        FRAME setlinewidth"
    print "        LLX 0.5 FRAME mul add LLY 0.5 FRAME mul add"
    print "        URX LLX sub FRAME sub URY LLY sub FRAME sub rectstroke"
    print "    grestore"
    print "} if"
    print ""
}


function set_gray()
{
    set_color("GRAY",GRAY,1,"setgray")
}


function set_hsb()
{
    set_color("HSB",HSB,3,"sethsbcolor")
}


function set_named_color( rgb,rgbfile,rgbline)
{
    if (RGBFILE == "")
    {
	RGBPATH = ENVIRON["RGBPATH"]
	if (RGBPATH == "")		# supply default fallback search path
	    RGBPATH = "/usr/lib/X11:/usr/openwin/lib/X11:/usr/local/lib/X11:/usr/local/X11R5/lib:/usr/lpp/X11/lib/X11"
	split(RGBPATH,rgbpath,":")
	rgbfile = find_file(rgbpath,"rgb.txt")
    }
    else
	rgbfile = RGBFILE
    if (rgbfile == "")
	warning("Cannot find X11 color mapping file [rgb.txt]")
    else
    {
	("grep 2>/dev/null -i '^[0-9 \t]*" BG "[ \t]*$' " rgbfile) | \
	    getline rgbline
	if (split(rgbline,rgb," ") > 3)
	    set_color("RGB",
	        (rgb[1] / 255) " " (rgb[2] / 255) " " (rgb[3] / 255),
	        3, "setrgbcolor")
	else
	    warning("named color [" BG "] not found in " rgbfile)
    }
}


function set_origin()
{
    OX = bigpoint((OX != "") ? OX : ((OXY != "") ? OXY : LLX))
    OY = bigpoint((OY != "") ? OY : ((OXY != "") ? OXY : LLY))
}


function set_rgb()
{
    set_color("RGB",RGB,3,"setrgbcolor")
}


function set_scale()
{
    if ((HEIGHT != "") || (WIDTH != "")) # absolute scaling
    {
	if (HEIGHT == "")	# uniform scaling
	    SX = SY = bigpoint(WIDTH)  / (URY - LLY)
	else if (WIDTH == "")	# uniform scaling
	    SX = SY = bigpoint(HEIGHT) / (URX - LLX)
	else			# nonuniform scaling
	{
	    SX = bigpoint(HEIGHT) / (URX - LLX)
	    SY = bigpoint(WIDTH)  / (URY - LLY)
	}
    }
    else			# relative scaling
    {
	SX = (SX != 0) ? SX : ((SXY != 0) ? SXY : 1)
	SY = (SY != 0) ? SY : ((SXY != 0) ? SXY : 1)
    }
}

function start_wrapper( k)
{
    set_origin()
    set_scale()
    do_header()

    print "BeginEPSF"
    print "%%BeginDocument:",((FILENAME == "") ? "stdin" : FILENAME)

    for (k = 1; k <= nlines; ++k) # print saved lines
	print lines[k]

    pre_scan = 0 # done with prescan: output() will now copy instead of save
}


function warning(message)
{
    print FILENAME ":" FNR ":%%" message >"/dev/stderr"
}
